/***********************************************************************\ 
*                                                                       * 
*   File: calendric_system/newh/calendar/gregorian/gregorian_aggregate.h
*                                                                       * 
*   The MultiCal System is free software in the public domain; you can  * 
*   redistribute it and/or modify it as you wish. We ask that you       * 
*   retain credits referencing the University of Arizona and that you   * 
*   identify any changes you make.                                      * 
*                                                                       * 
*   Report problems to multical@cs.arizona.edu                          * 
*   Direct all inquiries to:    The MultiCal Project                    * 
*                               Department of Computer Science          * 
*                               University of Arizona                   * 
*                               Tucson, AZ 85721                        * 
*                               U.S.A.                                  * 
*                                                                       * 
\***********************************************************************/ 
#ifndef gregorian_arithmetic
#define gregorian_arithmetic
/* aggregate.h
*
*  This module contains all function headers for variable
*  span aggregate computation.  Additional functions are
*  provided to convert partial variable span results to 
*  equivalent fixed span partial results.
*
*/

/*
*  Routine:  cal_standard_gregorian_v1_count_fstate_to_vstate
*
*  Description:  Convert a partial result of the count 
*		 of fixed spans to a count of variable spans
*
*  Arguments:  fs_count_state --  (IN) : the count of spans so far
*              vs_count_state --  (OUT): the count of fixed spans
*
*  Return Value: Error code
*
*  Errors: conversion_error
*
*  Side Effects: Overwrites vs_count_state
*/
error_type cal_standard_gregorian_v1_count_fstate_to_vstate(int *fs_count_state, void **vs_count_state);

/*
*  Routine:  cal_standard_gregorian_v1_accum_count_s
*
*  Description: This function increments the count by one
*
*  Arguments:   count_state -- (IN&OUT) : structure containing the count 
*				        of previously read in tuples
*               span        --     (IN) : the next span fed in
*
*  Return Value:  Error code
*
*  Errors: overflow
*
*  Side Effects: Overwrites count_state
*/
error_type cal_standard_gregorian_v1_accum_count_s(void *count_state, span_type span);

/*
*  Routine:  cal_standard_gregorian_v1_final_count_s
*
*  Description: The final count of spans
*
*  Arguments: final_state       -- (IN) : the final state contains the count
*             final_count_state -- (OUT): the count
*
*  Return Value:  Error code
*
*  Errors: None
*
*  Side Effects: Overwrites final_count_state
*/
error_type cal_standard_gregorian_v1_final_count_s(void *final_state, int *final_count);

/*
*  Routine:  cal_standard_gregorian_v1_sum_vstate_to_fstate
*
*  Description:  Convert a partial result of a sum of 
*		 variable spans to a sum of fixed spans
*
*  Arguments: fs_sum_state   --      (IN) : the sum of spans so far
*             vs_count_state --      (OUT): the sum of variable spans
*
*  Return Value:  Error code
*
*  Errors: conversion_error
*
*  Side Effects: Overwrites vs_sum_state
*/
error_type cal_standard_gregorian_v1_sum_fstate_to_vstate(span_type fs_sum_state, void **vs_sum_state);

/*
*  Routine:  cal_standard_gregorian_v1_accum_sum_s
*
*  Description: The sum of the spans read so far. The sum is calculated
*               by adding the input span and the previously accumulated
*               sum stored in sum_state, and then stored in sum_state
*
*  Arguments:   sum_state --   (IN&OUT) : previous state
*               span      --       (IN) : the span fed in
*
*  Return Value:  Error code
*
*  Errors: overflow
*
*  Side Effects: Overwrites sum_state
*/
error_type cal_standard_gregorian_v1_accum_sum_s(void *sum_state, span_type span);

/*
*  Routine:  cal_standard_gregorian_v1_final_sum_s
*
*  Description: The total sum of the spans. The sum is actually
*               stored in the input, final_state, but this function is
*               needed to return the final sum
*
*  Arguments: final_state --   (IN) : the final state
*             span        --   (OUT): the final result of sum of spans
*
*  Return Value:  Error code
*
*  Errors: None
*
*  Side Effects: Overwrites span
*/
error_type cal_standard_gregorian_v1_final_sum_s(void *final_state, span_type span);

/*
*  Routine:  cal_standard_gregorian_v1_avg_fstate_to_vstate
*
*  Description:  Convert a partial result of an average of 
*		 variable spans to an average of fixed spans
*
*  Arguments: fs_avg_state --    (IN) : the average of the spans so far
*             vs_avg_state --    (OUT): the average of fixed spans, 
*                                       converted to var state
*  Return Value: Error code
*
*  Errors: conversion_error
*
*  Side Effects: Overwrites vs_avg_state
*/
error_type cal_standard_gregorian_v1_avg_fstate_to_vstate(avg_state_fs fs_avg_state, void **vs_avg_state);

/*
*  Routine:  cal_standard_gregorian_v1_accum_avg_s
*
*  Description: Accumulate the sum of the spans read so far.
*               The new sum is calculated by adding the input span 
*		and the previously accumulated sum stored in 
*		avg_state. The sum in avg_state is then updated
*               with the new sum and stored.
*
*  Arguments:   avg_state --  (IN&OUT) : the previous state
*               span      --      (IN) : the span fed in
*
*  Return Value: Error code
*
*  Errors: overflow
*
*  Side Effects: Overwrites avg_state
*/
error_type cal_standard_gregorian_v1_accum_avg_s(void *avg_state, span_type span);

/*
*  Routine:  cal_standard_gregorian_v1_final_avg_s
*
*  Description: The final average of the spans. The average is
*               computed from the division of the sum and the count which
*               are stored in final_state (input argument)
*
*  Arguments: final_state -- (IN) : structure containing the total sum and count
*             span        -- (OUT): the final average of spans
*
*  Return Value: Error code
*
*  Errors: divide_by_zero
*
*  Side Effects: Overwrites span
*/
error_type cal_standard_gregorian_v1_final_avg_s(void *final_state, span_type span);

/*
*  Routine:  cal_standard_gregorian_v1_max_fstate_to_vstate
*
*  Description:  Convert a partial result of a maximum of variable spans 
*		 to an equivalent fixed span
*
*  Arguments: fs_max_state --  (IN) : the maximum span so far
*             vs_max_state --  (OUT): the maximum as a fixed span
*
*  Return Value: Error code
*
*  Errors: conversion_error
*
*  Side Effects: Overwrites vs_max_state
*/
error_type cal_standard_gregorian_v1_max_fstate_to_vstate(span_type fs_max_state, void **vs_max_state);


/*
*  Routine:  cal_standard_gregorian_v1_accum_max_s
*
*  Description: This function compares the input span with the maximum
*               stored in max_state. max_state is
*               set to the maximum of the new span and max_state
*
*  Arguments:   max_state --  (IN&OUT) : structure containing the maximum 
*                                       previously read span
*               span      --      (IN) : the new span fed in
*
*  Return Value: Error code
*
*  Errors: None
*
*  Side Effects: Overwrites max_state
*/
error_type cal_standard_gregorian_v1_accum_max_s(void *max_state, span_type span);

/*
*  Routine:  cal_standard_gregorian_v1_final_max_s
*
*  Description: The final result is the maximal span of a column
*
*  Arguments: final_state -- (IN) : final state contains the maximum of spans
*             span        -- (OUT): the result
*
*  Return Value: Error code
*
*  Errors: extremum_undefined
*
*  Side Effects: Overwrites span
*/
error_type cal_standard_gregorian_v1_final_max_s(void *final_state, span_type span);

/*
*  Routine:  cal_standard_gregorian_v1_min_fstate_to_vstate
*
*  Description:  Convert a partial result of a minimum of variable spans 
*		 to an equivalent fixed span
*
*  Arguments: fs_min_state --  (IN) : the minimum span so far
*             vs_min_state --  (OUT): the minimum as a fixed span
*
*  Return Value: Error code
*
*  Errors: conversion_error
*
*  Side Effects: Overwrites vs_min_state
*/
error_type cal_standard_gregorian_v1_min_fstate_to_vstate(span_type fs_min_state, void **vs_min_state);


/*
*  Routine:  cal_standard_gregorian_v1_accum_min_s
*
*  Description: This function compares the input span with the minimum
*               stored in min_state. min_state is set to the
*               minimum of min_state and span
*
*  Arguments:   min_state --   (IN&OUT) : structure containing the minimum 
*                                         previously read span
*               span      --       (IN) : the new span fed in
*
*  Return Value: Error code
*
*  Errors: None
*
*  Side Effects: Overwrites min_state
*/
error_type cal_standard_gregorian_v1_accum_min_s(void *min_state, span_type span);

/*
*  Routine:  cal_standard_gregorian_v1_final_min_s
*
*  Description: The final result is the minimum span of a column
*
*  Arguments: final_state -- (IN) : final state contains the minimum of spans
*             span        -- (OUT): the result
*
*  Return Value: Error code
*
*  Errors: extremum_undefined
*
*  Side Effects: Overwrites span
*/
error_type cal_standard_gregorian_v1_final_min_s(void *final_state, span_type span);
#endif
