/***********************************************************************\ 
*                                                                       * 
*   File: calendric_system/tadt/tadt_arithmetic.c
*                                                                       * 
*   The MultiCal System is free software in the public domain; you can  * 
*   redistribute it and/or modify it as you wish. We ask that you       * 
*   retain credits referencing the University of Arizona and that you   * 
*   identify any changes you make.                                      * 
*                                                                       * 
*   Report problems to multical@cs.arizona.edu                          * 
*   Direct all inquiries to:    The MultiCal Project                    * 
*                               Department of Computer Science          * 
*                               University of Arizona                   * 
*                               Tucson, AZ 85721                        * 
*                               U.S.A.                                  * 
*                                                                       * 
\***********************************************************************/ 
#include "system_wide.h"
#include "tadt.h"

/* Use the indeterminate basic adds, if indeterminacy is desired */
#ifdef INDETERMINATE
#define ADD(x,y,z) tadti_add((event_type) x, (span_type) y, (event_type) z);
#define SUB(x,y,z) tadti_sub((event_type) x, (span_type) y, (event_type) z);
static error_type tadti_add(event_type e, span_type s, event_type r);
static error_type tadti_sub(event_type e, span_type s, event_type r);
#else /*DETERMINATE */
#define ADD(x,y,z) tadt_add((event_type) x, (span_type) y, (event_type) z);
#define SUB(x,y,z) tadt_add((event_type) x, (span_type) y, (event_type) z);
#endif 
static error_type tadt_add(event_type e, span_type s, event_type r);
static error_type tadt_sub(event_type e, span_type s, event_type r);


/*  
*  Routine:  tadt_neg_fs
*
*  Description: Determine the sign opposite fixed span which is otherwise the
*		same as the input span.  Has no effect on a zero length
*               span.
*
*  Arguments: 	span -- (IN) : span to be converted
*               result -- (OUT) : result span 
*
*  Return Value: Error code
*
*  Errors: out_of_range (Max negative span has no positive counterpart)
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_neg_fs(span_type span, span_type result) {
  error_type rc = tadt_ok;
  int32 temp32;
  int64 temp64;
  int96 temp96;
  
  result->upack.modified = TRUE;
  switch (span->generic.ts_type) {
    case upack32:
      result->upack.ts_type = upack32;
      LOAD32(-1,temp32);
      MULT32(temp32,span->upack32.granules,result->upack32.granules); 
      return tadt_ok;
      break;

    case upack64: 
      result->upack.ts_type = upack64;
      LOAD64(-1,temp64); 
      MULT64(temp64,span->upack64.granules,result->upack64.granules); 
      return tadt_ok;
      break;

    case upack96: 
      LOAD96(-1,temp96);
      MULT96(temp96,span->upack96.granules,result->upack96.granules); 
      return tadt_ok;
      break;

    case special:
      result->special.ts_type = special;
      switch (span->special.which) {
        case all_of_time: 
           result->special.which = negative_all_of_time; break;
        case negative_all_of_time: 
           result->special.which = all_of_time; break;
        case undefined:
           result->special.which = undefined; break;
        default: 
           internal_error("%s", "tadt_neg_fs: Unknown special span type");
      }
 
    default:
      internal_error("%s", "tadt_neg_fs: bad ts_type");

    }

  return ok; 
}

/*  
*  Routine:  tadt_fs_add_fs
*
*  Description: Add two fixed spans
*
*  Arguments: 	span1, span2 -- (IN) : spans to be added
*               result -- (OUT) : result span
*
*  Return Value: Error code
*
*  Errors: out_of_range
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_fs_add_fs(span_type span1, span_type span2, span_type result) {

  switch (span1->upack.ts_type << 4) {

  + span2->upack.ts_type) {
    case upack32 * 16 + upack32:    /* upack32 and upack32*/
      ADD32_TO_32(span1, span2, result);
      return RC;

    case upack32 * 16 + upack32:    /* upack32 and upack32*/
      ADD32_TO_32(span1, span2, result);
      return RC;

    case upack32 * 16 + special: /* upack and special */
    case upack64 * 16 + special: 
    case upack96 * 16 + special: 
      tadt_copy_s(span2, result);
      return ok;

    case special * 16 + upack32: /* special and upack */
    case special * 16 + upack64: 
    case special * 16 + upack96: 
      tadt_copy_s(span1, result);
      return ok;

    case special * 16 + special: /* special and special */
      result->special.ts_type = special;
      switch (span1->special.which) {
        case all_of_time: 
          switch (span2->special.which) {
            case all_of_time: 
              result->special.which = all_of_time; return ok;
            case negative_all_of_time: 
	      result->upack.ts_type = upack32; 
	      /* result->upack.precision = {10,10,0}; */
              result->upack.modified = TRUE;
              LOAD32(0,result->upack32.granules); 
	      return ok;
            case undefined:
              result->special.which = undefined; return ok;
            default: internal_error("%s", "tadt_fs_add_fs: Bad special type");
            }
        case negative_all_of_time: 
          switch (span2->special.which) {
            case all_of_time: 
	      result->upack.ts_type = upack32; 
              result->upack.modified = TRUE;
	      /* result->upack.precision = {10,10,0}; */
              result->upack.modified = TRUE;
              LOAD32(0,result->upack32.granules); 
	      return ok;
            case negative_all_of_time: 
              result->special.which = negative_all_of_time; return ok;
            case undefined:
              result->special.which = undefined; return ok;
            default: internal_error("%s", "tadt_fs_add_fs: Bad special type");
            }
        case undefined: 
          result->special.which = undefined; return ok;
        default: internal_error("%s", "tadt_fs_add_fs: Bad special type");
        }

    default:
      internal_error("%s", "tadt_fs_add_fs: BAD Flags");

  }

  internal_error("%s", "tadt_fs_add_fs: how did I get here?");
  return ok; /* should never get here */
}


/*  
*  Routine:  tadt_fs_minus_fs
*
*  Description: Subtract two fixed spans
*
*  Arguments: 	span1 -- (IN) : minuend (span to be subtracted from)
*		span2 -- (IN) : subtrahend
*               result -- (OUT) : result span
*
*  Return Value: Error code
*
*  Errors: out_of_range
*          One note: a loss of resolution is flagged as out of range.
*           (e.g., a hires + lowres --> hires or out_of_range)
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_fs_minus_fs(span_type span1, span_type span2, span_type result){
  switch ((span1->generic.ts_type << 4) + span2->generic.ts_type) {

    case upack * 16 + upack:    /* upack and upack*/
      return SUB(span1, span2, result);

    case upack * 16 + special: /* upack and special */
      result->special.ts_type = special;
      switch (span2->special.which) {
        case all_of_time: 
	  result->special.which = negative_all_of_time; return ok;
        case negative_all_of_time: 
	  result->special.which = all_of_time; return ok;
        case undefined: 
	  result->special.which = undefined; return ok;
        default: internal_error("%s", "tadt_fs_minus_fs: Bad special type");
	}

    case special * 16 + upack: /* special and upack */
      tadt_copy_s(span1, result);
      return ok;

    case special * 16 + special: /* special and special */
      result->special.ts_type = special;
      switch (span1->special.which) {
        case all_of_time: 
          switch (span2->special.which) {
            case all_of_time: 
	      result->upack.ts_type = upack; 
              result->upack.size = 32;
              result->upack.modified = TRUE;
	      /* result->upack.precision = {10,10,0}; */
              result->upack.modified = TRUE;
              LOAD32(0,result->upack32.granules); 
	      return ok;
            case negative_all_of_time: 
              result->special.which = all_of_time; return ok;
            case undefined:
              result->special.which = undefined; return ok;
            default: internal_error("%s", "tadt_fs_minus_fs: Bad special type");
            }
        case negative_all_of_time: 
          switch (span2->special.which) {
            case all_of_time: 
              result->special.which = negative_all_of_time; return ok;
            case negative_all_of_time: 
	      result->upack.ts_type = upack; 
              result->upack.size = 32;
              result->upack.modified = TRUE;
	      /* result->upack.precision = {10,10,0}; */
              result->upack.modified = TRUE;
              LOAD32(0,result->upack32.granules); 
	      return ok;
            case undefined:
              result->special.which = undefined; return ok;
            default: internal_error("%s", "tadt_fs_minus_fs: Bad special type");
            }
        case undefined: 
          result->special.which = undefined; return ok;
        default: internal_error("%s", "tadt_fs_minus_fs: Bad special type");
        }

    default:
      internal_error("%s", "tadt_fs_minus_fs: BAD Flags");

  }

  internal_error("%s", "tadt_fs_minus_fs: how did I get here?");
  return ok; /* should never get here */
}


/*  
*  Routine:  tadt_e_add_fs
*
*  Description: Displace an event by adding a span.  Result is always
*               an extended resolution timestamp except for adding a
*               lowres to lowres.  
*
*  Arguments: event -- (IN) : the event time
*	      span -- (IN) : the displacement
*             result -- (OUT) : result event
*
*  Return Value: Error code
*
*  Errors: out_of_range
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_e_add_fs(event_type event, span_type span, event_type result) {

  switch ((event->generic.ts_type << 4) + span->generic.ts_type) {

    case upack * 16 + upack:    /* upack and upack*/
      return ADD( event, span, result);

    case upack * 16 + special: /* upack and special */
      switch (span->special.which) {
	case all_of_time: 
           result->special.which = forever;
           result->special.ts_type = special;
           return ok;

	case negative_all_of_time:  
           result->special.which = beginning;
           result->special.ts_type = special;
           return ok;

	case undefined: 
           result->special.which = undefined;
           result->special.ts_type = special;
           return ok;

        default:
           internal_error("%s", "tadt_e_add_fs: bad special type");
	}

    case special * 16 + upack: /* special and upack */
      tadt_copy_e(event, result);
      return ok;

    case special * 16 + special: /* special and special */
      result->special.ts_type = special;
      switch (event->special.which) {
        case beginning: 
        case forever: 
          switch (span->special.which) {
            case all_of_time: 
              result->special.which = forever; return ok;
            case negative_all_of_time: 
              result->special.which = beginning; return ok;
            case undefined:
              result->special.which = undefined; return ok;
            default: internal_error("%s", "tadt_e_add_fs: Bad special type");
            }
        case undefined: 
          result->special.which = undefined; return ok;
        default: internal_error("%s", "tadt_e_add_fs: Bad special type");
        }

    default:
      internal_error("%s", "tadt_e_add_fs: BAD Flags, must unpack event or span");

  }

  internal_error("%s", "tadt_e_add_fs: how did I get here?");
  return ok; /* should never get here */
}

/*  
*  Routine:  tadt_e_minus_fs
*
*  Description: Displace an event by subtracting a span 
*
*  Arguments: event -- (IN) : the event time
*	      span  -- (IN) : the displacement
*             result -- (OUT) : result event
*
*  Return Value: Error code
*
*  Errors: out_of_range
*          overflow  (beginning - anything)
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_e_minus_fs(event_type event, span_type span, event_type result) {
  switch ((event->generic.ts_type << 4) + span->generic.ts_type) {

    case upack * 16 + upack:    /* upack and upack*/
      return SUB(event, span, result);

    case upack * 16 + special: /* upack and special */
      switch (span->special.which) {
	case all_of_time: 
           result->special.which = beginning;
           result->special.ts_type = special;
           return ok;

	case negative_all_of_time:  
           result->special.which = forever;
           result->special.ts_type = special;
           return ok;

	case undefined: 
           result->special.which = undefined;
           result->special.ts_type = special;
           return ok;

        default:
           internal_error("%s", "tadt_e_minus_fs: bad special type");
	}

    case special * 16 + upack: /* special and upack */
      tadt_copy_e(event, result);
      return ok;

    case special * 16 + special: /* special and special */
      result->special.ts_type = special;
      switch (event->special.which) {
        case beginning: 
        case forever: 
          switch (span->special.which) {
            case all_of_time: 
              result->special.which = beginning; return ok;
            case negative_all_of_time: 
              result->special.which = forever; return ok;
            case undefined:
              result->special.which = undefined; return ok;
            default: internal_error("%s", "tadt_e_minus_fs: Bad special type");
            }
        case undefined: 
          result->special.which = undefined; return ok;
        default: internal_error("%s", "tadt_e_minus_fs: Bad special type");
        }

    default:
      internal_error("%s", "tadt_e_minus_fs: BAD Flags");

  }

  internal_error("%s", "tadt_e_minus_fs: how did I get here?");
  return ok; /* should never get here */
}

/*  
*  Routine:  tadt_add
*
*  Description: Do the addition of an event and a span.  Favor a positive
*               result.
*
*  Arguments:   event, span -- (IN) : input event and span
*               result -- (OUT) : result event
*
*  Return Value: Error code
*
*  Errors: out_of_range
*  
*  Side Effects: Assignment to passed result structure
*/

static error_type tadt_add(event_type event, span_type span, event_type result){
  error_type rc = tadt_ok;

  result->generic.ts_type = upack;
  result->upack.modified = TRUE;
  result->upack.size = event->upack.size;
  switch (event->upack.size) {
    case SIZE32:
      rc = ADD32(event->upack32.granules,span->upack32.granules,result->upack32.granules); break;
    case SIZE64:
      rc = ADD64(event->upack64.granules,span->upack64.granules,result->upack64.granules); break;
    case SIZE96:
      rc = ADD96(event->upack96.granules,span->upack96.granules,result->upack96.granules); break;
    default: internal_error("%s", "bad size in tadt_arithmetic");
    }
  return rc;
}

/*  
*  Routine:  tadt_sub
*
*  Description: Do the subtraction of an event and a span.  Favor a positive
*               result.
*
*  Arguments:   event, span -- (IN) : input event and span
*               result -- (OUT) : result event
*
*  Return Value: Error code
*
*  Errors: out_of_range
*  
*  Side Effects: Assignment to passed result structure
*/
static error_type tadt_sub(event_type event, span_type span, event_type result){
  error_type rc = tadt_ok;

  result->generic.ts_type = upack;
  result->upack.modified = TRUE;
  result->upack.size = event->upack.size;
  switch (event->upack.size) {
    case SIZE32:
      rc = SUB32(event->upack32.granules,span->upack32.granules,result->upack32.granules); break;
    case SIZE64:
      rc = SUB64(event->upack64.granules,span->upack64.granules,result->upack64.granules); break;
    case SIZE96:
      rc = SUB96(event->upack96.granules,span->upack96.granules,result->upack96.granules); break;
    default: internal_error("%s", "bad size in tadt_arithmetic");
    }

  return rc;

}

#ifdef INDETERMINATE

/*
*  Routine:  tadti_add
*
*  Description: Do the addition of an event and a span, either of which
*               could be indeterminate.
*
*  Arguments:   event, span -- (IN) : input event and span
*               result -- (OUT) : result event
*
*  Return Value: Error code
*
*  Errors: out_of_range
*
*  Side Effects: Assignment to passed result structure
*/

static error_type tadti_add(event_type event, span_type span, event_type result)
{
  error_type rc;

  if (event->upack.indeterminate) {
    if (span->upack.indeterminate) {

      /* Add an indeterminate event and an indeterminate span */
      rc = tadt_add( (event_type) &event->upack_ind.lower,
                     (span_type) &span->upack_ind.lower,
                     (event_type) &result->upack_ind.lower);
      if (rc) return rc;
      rc = tadt_add( (event_type) &event->upack_ind.upper,
                     (span_type) &span->upack_ind.upper,
                     (event_type) &result->upack_ind.upper);
      result->upack_ind.indeterminate = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      return rc;
      }
    else {

      /* Add an indeterminate event and a determinate span */
      rc = tadt_add( (event_type) &event->upack_ind.lower,
                     span,
                     (event_type) &result->upack_ind.lower);
      if (rc) return rc;
      rc = tadt_add( (event_type) &event->upack_ind.upper,
                     span,
                     (event_type) &result->upack_ind.upper);
      result->upack_ind.indeterminate = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      return rc;
      }
    }
  else {
    if (span->upack.indeterminate) {

      /* Add a determinate event and an indeterminate span */
      rc = tadt_add( event,
                     (span_type) &span->upack_ind.lower,
                     (event_type) &result->upack_ind.lower);
      if (rc) return rc;
      rc = tadt_add( event,
                     (span_type) &span->upack_ind.upper,
                     (event_type) &result->upack_ind.upper);
      result->upack_ind.indeterminate = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      return rc;
      }
    else {

      /* Add a determinate event and a determinate span */

      return tadt_add(event, span, result);
      }
    }

}


/*
*  Routine:  tadti_sub
*
*  Description: Do the subtraction of a span from an event, either of which
*               could be indeterminate.
*
*  Arguments:   event, span -- (IN) : input event and span
*               result -- (OUT) : result event
*
*  Return Value: Error code
*
*  Errors: out_of_range
*
*  Side Effects: Assignment to passed result structure
*/

static error_type tadti_sub(event_type event, span_type span, event_type result)
{
  error_type rc;

  if (event->upack.indeterminate) {
    if (span->upack.indeterminate) {

      /* Add an indeterminate event and an indeterminate span */
      rc = tadt_sub( (event_type) &event->upack_ind.lower,
                     (span_type) &span->upack_ind.lower,
                     (event_type) &result->upack_ind.lower);
      if (rc) return rc;
      rc = tadt_sub( (event_type) &event->upack_ind.upper,
                     (span_type) &span->upack_ind.upper,
                     (event_type) &result->upack_ind.upper);
      result->upack_ind.indeterminate = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      return rc;
      }
    else {

      /* Subtract an indeterminate event and a determinate span */
      rc = tadt_sub( (event_type) &event->upack_ind.lower,
                     span,
                     (event_type) &result->upack_ind.lower);
      if (rc) return rc;
      rc = tadt_sub( (event_type) &event->upack_ind.upper,
                     span,
                     (event_type) &result->upack_ind.upper);
      result->upack_ind.indeterminate = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      return rc;
      }
    }
  else {
    if (span->upack.indeterminate) {

      /* Subtract a determinate event from an indeterminate span */
      rc = tadt_sub( event,
                     (span_type) &span->upack_ind.lower,
                     (event_type) &result->upack_ind.lower);
      if (rc) return rc;
      rc = tadt_sub( event,
                     (span_type) &span->upack_ind.upper,
                     (event_type) &result->upack_ind.upper);
      result->upack_ind.indeterminate = TRUE;
      result->upack_ind.distribution.name = 0/*unknown distribution*/;
      return rc;
      }
    else {

      /* Add a determinate event and a determinate span */
      return tadt_sub(event, span, result);
      }
    }
}
#endif

/*  
*  Routine:  tadt_e_minus_e
*
*  Description: Compute the span between two events by 
*		subtracting the second event from the first event (e1 - e2).
*		if e1 > e2,  return a positive span
*		if e1 < e2,  return a negative span
*
*  Arguments:   event1, event2 -- (IN) : two input events.
*               result -- (OUT) : result span
*
*  Return Value: Error code
*
*  Errors: none
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_e_minus_e(event_type event1, event_type event2, span_type result) {
  switch ((event1->generic.ts_type << 4) + event2->generic.ts_type) {

    case upack * 16 + upack:    /* upack and upack*/
      return SUB(event1, event2, result);

    case upack * 16 + special: /* upack and special */
      switch (event2->special.which) {
	case forever: 
           result->special.which = all_of_time;
           result->special.ts_type = special;
           return ok;

	case beginning:  
           result->special.which = negative_all_of_time;
           result->special.ts_type = special;
           return ok;

	case undefined: 
           result->special.which = undefined;
           result->special.ts_type = special;
           return ok;

        default:
           internal_error("%s", "tadt_e_minus_e: bad special type");
	}

    case special * 16 + upack: /* special and upack */
      tadt_copy_e(event1, (event_type) result);
      return ok;

    case special * 16 + special: /* special and special */
      result->special.ts_type = special;
      switch (event1->special.which) {
        case beginning: 
          switch (event2->special.which) {
            case beginning: 
	      result->upack.ts_type = upack; 
	      result->upack.modified = TRUE; 
	      result->upack32.granules = 0;
	      return ok;
            case forever: 
              result->special.which = negative_all_of_time; return ok;
            case undefined:
              result->special.which = undefined; return ok;
            default: internal_error("%s", "tadt_e_minus_e: Bad special type");
            }
        case forever: 
          switch (event2->special.which) {
            case beginning: 
              result->special.which = all_of_time; return ok;
            case forever: 
	      result->upack.ts_type = upack; 
	      result->upack.modified = TRUE; 
	      result->upack32.granules = 0;
	      return ok;
            case undefined:
              result->special.which = undefined; return ok;
            default: internal_error("%s", "tadt_e_minus_e: Bad special type");
            }

        case undefined: 
          result->special.which = undefined; return ok;
        default: internal_error("%s", "tadt_e_minus_e: Bad special type");
        }

    default:
      internal_error("%s", "tadt_e_minus_e: BAD Flags");

  }

  internal_error("%s", "tadt_e_minus_e: how did I get here?");
  return ok; /* should never get here */
}

/*  
*  Routine:  tadt_fs_times_n
*
*  Description: Return a span equal to a multiple of the input span.  The
*               multiplier must be an integer.
*
*  Arguments:   span -- (IN): the input span
*		n -- (IN): number of times
*               result -- (OUT) : result span
*
*  Return Value: Error code
*
*  Errors: out_of_range
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_fs_times_n(span_type span, int n, span_type result) {
  int32 temp32;
  int64 temp64;
  int96 temp96;
  error_type rc = tadt_ok; 

  result->upack.modified = TRUE;
  switch (span->generic.ts_type) {
    case upack: 
      result->upack.ts_type = upack;
      result->upack.size = span->upack.size;
      switch (span->upack.size) {
        case SIZE32: 
          LOAD32(n,temp32);
          MULT32(temp32,span->upack32.granules,result->upack32.granules); 
	  rc = tadt_ok;
          break;
        case SIZE64: 
          LOAD64(n,temp64); 
	  MULT64(temp64,span->upack64.granules,result->upack64.granules); 
	  rc = tadt_ok;
	  break;
        case SIZE96: 
          LOAD96(n,temp96);
	  MULT96(temp96,span->upack96.granules,result->upack96.granules); 
	  rc = tadt_ok;
	  break;
        default: internal_error("%s", "bad size in tadt_arithmetic");
        }
      return rc;

    case special:
      result->special.ts_type = special;
      switch (span->special.which) {
        case all_of_time:
          result->special.which = (n < 0)? negative_all_of_time: all_of_time;
          break;
        case negative_all_of_time:
          result->special.which = (n < 0)? all_of_time: negative_all_of_time;
          break;
        case undefined:
          result->special.which = undefined;
          break;
        default: internal_error("%s", "tadt_fs_times_n: Bad special type");
        }
    default: internal_error("%s", "tadt_fs_times_n: Bad ts_type");
    }
  return ok;
}

/*  
*  Routine:  tadt_fs_div_n
*
*  Description: Divide a span by a numeric value.  N must be nonzero.
*
*  Arguments: span -- (IN):the input span
*	      n -- (IN): number of division
*             result -- (OUT) : result span
*
*  Return Value: Error code
*
*  Errors: out_of_range
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_fs_div_n(span_type span, int n, span_type result) {
  int32 temp32;
  int64 temp64;
  int96 temp96;
  error_type rc = tadt_ok; 
  
  result->upack.modified = TRUE;
  switch (span->generic.ts_type) {
    case upack: 
      result->upack.ts_type = upack;

      /* Special case division by zero */
      if (n == 0) {
	tadt_copy_s(span, result);
	return divide_by_zero;
	}

      result->upack.size = span->upack.size;
      switch (span->upack.size) {
        case SIZE32: 
          LOAD32(n,temp32);
          DIV32(span->upack32.granules,temp32,result->upack32.granules); 
	  rc = tadt_ok;
          break;
        case SIZE64: 
          LOAD64(n,temp64); 
	  DIV64(span->upack64.granules,temp64,result->upack32.granules); 
	  rc = tadt_ok;
	  break;
        case SIZE96: 
          LOAD96(n,temp96);
	  DIV96(span->upack96.granules,temp96,result->upack32.granules); 
	  rc = tadt_ok;
	  break;
        default: internal_error("%s", "bad size in tadt_arithmetic");
        }

      return rc;
       
    case special:
      result->special.ts_type = special;
      switch (span->special.which) {
        case all_of_time:
          result->special.which = (n < 0)? negative_all_of_time: all_of_time;
          break;
        case negative_all_of_time:
          result->special.which = (n < 0)? all_of_time: negative_all_of_time;
          break;
        case undefined:
          result->special.which = undefined;
          break;
        default: internal_error("%s", "tadt_fs_div_n: Bad special type");
        }
    default: internal_error("%s", "tadt_fs_div_n: Bad ts_type");
    }
  return ok;
}

/*  
*  Routine:  tadt_fs_div_fs
*
*  Description: Divide one span by another.  Because of limitations in the
*               hardware (only 64-bit operations) if the seconds field is
*               quite large, we ignore the fields with finer granularities
*               (e.g., milliseconds and microseconds).  We admit that this
*               is a hack, but those fields are unimportant in such a 
*               division.
*
*  Arguments:   span1 -- (IN): the dividend
*	  	span2 --(IN): the divisor
*               result -- (OUT) : result integer (double)
*
*  Return Value: Error code
*
*  Errors: division_by_zero
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_fs_div_fs(span_type span1, span_type span2, double *result){
  int32 temp32;
  int64 temp64;
  int96 temp96;
  error_type rc = tadt_ok; 
  
  switch ((span1->generic.ts_type << 4) + span2->generic.ts_type) {

    case upack * 16 + upack:    /* upack and upack*/

      switch (span2->upack.size) {
        case SIZE32: 
          LOAD32(0,temp32);
	  if (EQ32(temp32,span2->upack32.granules)) return divide_by_zero;
          rc = DIV32(span1->upack32.granules,span2->upack32.granules,*result); 
          break;
        case SIZE64: 
          LOAD64(0,temp64);
	  if (EQ64(temp64,span2->upack64.granules)) return divide_by_zero;
          rc = DIV64(span1->upack64.granules,span2->upack64.granules,*result); 
	  break;
        case SIZE96: 
          LOAD96(0,temp96);
	  if (EQ96(temp96,span2->upack96.granules)) return divide_by_zero;
          rc = DIV96(span1->upack96.granules,span2->upack96.granules,*result); 
	  break;
        default: internal_error("%s", "bad size in tadt_arithmetic");
        }

      return rc;

    case upack * 16 + special: /* upack and special */
      *result = 0;
      return ok;

    case special * 16 + upack: /* special and upack */
      return out_of_range;

    case special * 16 + special: /* special and special */
      switch (span1->special.which) {
        case all_of_time: 
          switch (span2->special.which) {
            case all_of_time: *result = 1; return ok;
            case negative_all_of_time: *result = -1; return ok;
            default: internal_error("%s", "tadt_fs_div_fs: Bad special type");
            }
        case negative_all_of_time: 
          switch (span2->special.which) {
            case all_of_time: *result = -1; return ok;
            case negative_all_of_time: *result = 1; return ok;
            default: internal_error("%s", "tadt_fs_div_fs: Bad special type");
            }
        default: internal_error("%s", "tadt_fs_div_fs: Bad special type");
        }

    default:
      internal_error("%s", "tadt_fs_div_fs: BAD Flags");

  }

  internal_error("%s", "tadt_fs_div_fs: how did I get here?");
  return ok; /* should never get here */
}

/*  
*  Routine:  tadt_i_add_fs
*
*  Description: Displace an interval by adding a span
*
*  Arguments: interval -- (IN): the interval to be displaced
*	      span -- (IN): the span of displacement
*             result -- (OUT) : result interval
*
*  Return Value: Error code
*
*  Errors: out_of_range
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_i_add_fs(interval_type interval, span_type span, interval_type result) {
  error_type rc;

  /* Displace each bounding event */
  if ((rc = tadt_e_add_fs(interval->start, span, result->start)) != ok)
     return rc;
  return tadt_e_add_fs(interval->terminate, span, result->terminate);
}

/*  
*  Routine:  tadt_i_minus_fs
*
*  Description: Displace an interval by subtracting a span
*
*  Arguments: interval -- (IN): the interval to be displaced
*	      span -- (IN): the span of displacement
*             result -- (OUT) : result interval
*
*  Return Value: Error code
*
*  Errors: out_of_range
*          overflow     (|beginning,forever| - anything)
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_i_minus_fs(interval_type interval, span_type span, interval_type result) {
  error_type rc;

  /* Displace each endpoint separately */
  if ((rc = tadt_e_minus_fs(interval->start, span, result->start)) != ok) return rc;
  return tadt_e_minus_fs(interval->terminate, span, result->terminate);
}  
