/***********************************************************************\ 
*                                                                       * 
*   File: calendric_system/tadt/tadt_timestamp.c
*                                                                       * 
*   The MultiCal System is free software in the public domain; you can  * 
*   redistribute it and/or modify it as you wish. We ask that you       * 
*   retain credits referencing the University of Arizona and that you   * 
*   identify any changes you make.                                      * 
*                                                                       * 
*   Report problems to multical@cs.arizona.edu                          * 
*   Direct all inquiries to:    The MultiCal Project                    * 
*                               Department of Computer Science          * 
*                               University of Arizona                   * 
*                               Tucson, AZ 85721                        * 
*                               U.S.A.                                  * 
*                                                                       * 
\***********************************************************************/ 
#include <stdlib.h>
#include "system_wide.h"
#include "tadt.h"

extern resolution_type combined_flags[MAX_RESOLUTIONS_OF_EVENTS][MAX_RESOLUTIONS_OF_EVENTS];
extern size_type sizes[MAX_RESOLUTIONS_OF_EVENTS][MAX_PRECISION];
extern context_type context;

#ifdef INDETERMINATE
extern dist_type distributions[MAX_DISTRIBUTION];
#endif

/*  
*  Routine:  tadt_unswizzle_e  NEEDS WORK
*
*  Description:  Packs an event timestamp from the swizzled format.
*                The size of target event is passed in.  If the event 
*                is too big an error is generated.  An event packs
*                to whatever kind it started out as.
*
*  Arguments: event -- (IN): event structure
*             size  -- (IN): size of result 
*             result -- (OUT): result event
*
*  Return Value: Error code
*
*  Errors: tadt_not_enough_space
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_unswizzle_e(event_type event, int size, char *result)
{
  packed_event_type res;

  res = (packed_event_type) result;
  switch (event->upack.ts_type) {
#ifdef NOWRELATIVE
    case nowrelative:
#endif
    case special: /* do not unswizzle, it is never unpacked */
      break;

    case upack:
      /* Implicit coercion? */
      res->generic.ts_type = event->upack.base_ts_type;
      break;

    default:
      internal_error("%s", "tadt_unswizzle_e: bad passed ts_type");

    }

  return ok;
}

/*
*  Routine:  tadt_unswizzle_fs
*
*  Description:  Packs a span timestamp from the swizzled format.
*                The size of target span is passed in.  If the span
*                is too big an error is generated.  A span packs
*                to whatever kind it started out as.
*
*  Arguments: span -- (IN): span structure
*             size  -- (IN): size of result 
*             result -- (OUT): resulting string of characters
*
*  Return Value: Error code
*
*  Errors: tadt_not_enough_space
*
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_unswizzle_fs(span_type span, int size, char *result) {
   /* exploit event, fixed span similarity */
   tadt_unswizzle_e((event_type) span, size, result);
   return ok;
}

/*
*  Routine:  tadt_unswizzle_i  NEEDS WORK
*
*  Description:  Packs an interval timestamp from the swizzled format.
*                The size of target interval is passed in.  If the interval
*                is too big an error is generated.  A interval packs
*                to whatever kind it started out as.
*
*  Arguments: interval -- (IN): interval structure
*             size  -- (IN): size of result 
*             result -- (OUT): resulting string of chars
*
*  Return Value: Error code
*
*  Errors: tadt_not_enough_space
*
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_unswizzle_i(interval_type interval, int size, char *result) {
   /* Assume that half the thing is in each place? */
   tadt_unswizzle_e(interval->start, size, result);
   tadt_unswizzle_e(interval->terminate, size, result + (size >> 1));
   return ok;
}

#ifdef INDETERMINATE

/* This is the chunk table */
extern int32 chunk32_table[HOW_MANY_CHUNKS][MAX_CHUNK];
extern int64 chunk64_table[HOW_MANY_CHUNKS][MAX_CHUNK];
extern int96 chunk96_table[HOW_MANY_CHUNKS][MAX_CHUNK];

/* An easy way to specify a single chunk */
#define CHUNK32(field,var) \
   chunk32_table[var->field.chunks][var->field.chunksize]

#define CHUNK64(field,var) \
   chunk64_table[var->field.chunks][var->field.chunksize]

#define CHUNK96(field,var) \
   chunk96_table[var->field.chunks][var->field.chunksize]

/*  
*  Routine:  tadt_unpack_chunked
*
*  Description:  Unpacks a chunked timestamp with a standard distribution.
*
*  Arguments: event -- (IN): event structure
*             result -- (OUT): result event
*
*  Return Value: Error code
*
*  Errors: none
*  
*  Side Effects: Assignment to passed result structure
*/

static error_type tadt_unpack_chunked(packed_event_type event, event_type result, precision_type precision)
{
  long temp;
  size_type size;
  int32 temp32;
  int64 temp64;
  int96 temp96;
  event_type result_lower, result_upper;

  /* PRECISION_TO_SIZE(precision,size); */
  size = sizes[packed_chunked][precision->precision];

  result->upack.ts_type = upackind;
  result->upack_ind.base_ts_type = packed_indeterminate;
  result->upack_ind.modified = FALSE;

  result_lower = (event_type) &result->upack_ind.lower;
  result_lower->upack.ts_type = size;
  result_lower->upack.base_ts_type = undefined;
  result_lower->upack.modified = FALSE;
  result_upper = (event_type) &result->upack_ind.upper;
  result_upper->upack.ts_type = size;
  result_upper->upack.base_ts_type = undefined;
  result_upper->upack.modified = FALSE;

  result->upack_ind.distribution.right_off = 0;
  result->upack_ind.distribution.left_off = 0;
  result->upack_ind.distribution.name = (event->chunked.missing)? MISSING: UNIFORM;

  /* Figure out chunk in terms of granules */
  temp = event->word[1];
  if (event->chunked.sign) {
    temp = temp | CHUNKED_STANDARD_MASK;
    }

  switch (size) {
    case upack32: 
      CONSTRUCT32(temp,temp32);
      ASSIGN32(temp32,result_lower->upack32.granules);
      ADD32(result_lower->upack32.granules, 
	    CHUNK32(chunked,event), 
	    result_upper->upack32.granules);
      /*
      tadt_e_add_fs(result_lower,
 		    CHUNK32(chunked,event), 
		    result_upper);
      */
      break;
    case upack64: 
      CONSTRUCT64(temp,event->word[2],temp64);
      ASSIGN64(temp64,result_lower->upack64.granules);
      ADD64(result_lower->upack64.granules, 
	    CHUNK64(chunked,event), 
	    result_upper->upack64.granules);
      /*
      tadt_e_add_fs(result_lower, 
 		    CHUNK64(chunked,event), 
		    result_upper);
      */
      break;
    case upack96: 
      CONSTRUCT96(temp,event->word[2],event->word[3],temp96);
      ASSIGN96(temp96,result_lower->upack96.granules);
      ADD96(result_lower->upack96.granules, 
	    CHUNK96(chunked,event), 
	    result_upper->upack96.granules);
      /*
      tadt_e_add_fs(result_lower, 
 		    CHUNK96(chunked,event), 
		    result_upper);
      */
      break;
    default:
      internal_error("%s", "bad size");
    }
  return tadt_ok;
}

/*  
*  Routine:  tadt_unpack_chunked_nonstandard NEEDS WORK
*
*  Description:  Unpacks a chunked timestamp with a nonstandard distribution.
*
*  Arguments: event -- (IN): event structure
*             result -- (OUT): result event
*             precision -- (IN): precision info
*
*  Return Value: Error code
*
*  Errors: none
*  
*  Side Effects: Assignment to passed result structure
*/

static error_type tadt_unpack_chunked_nonstandard(packed_event_type event, event_type result, precision_type precision)
{
 return tadt_ok;
}

/*  
*  Routine:  tadt_unpack_general NEEDS WORK
*
*  Description:  Unpacks a general timestamp with a standard distribution.
*
*  Arguments: event -- (IN): event structure
*             result -- (OUT): result event
*             precision -- (IN): precision info
*
*  Return Value: Error code
*
*  Errors: none
*  
*  Side Effects: Assignment to passed result structure
*/

static error_type tadt_unpack_general(packed_event_type event, event_type result, precision_type precision)
{
 return tadt_ok;
}

/*  
*  Routine:  tadt_unpack_general_nonstandard NEEDS WORK
*
*  Description:  Unpacks a general timestamp with a nonstandard distribution.
*
*  Arguments: event -- (IN): event structure
*             result -- (OUT): result event
*             precision -- (IN): precision info
*
*  Return Value: Error code
*
*  Errors: none
*  
*  Side Effects: Assignment to passed result structure
*/

static error_type tadt_unpack_general_nonstandard(packed_event_type event, event_type result, precision_type precision)
{
 return tadt_ok;
}

#endif

/*  
*  Routine:  tadt_unpack_determinate
*
*  Description:  Unpacks a basic determinate timestamp into the unpacked format.
*
*  Arguments: event -- (IN): event structure
*             result -- (OUT): result event
*
*  Return Value: Error code
*
*  Errors: none
*  
*  Side Effects: Assignment to passed result structure
*/

static error_type tadt_unpack_determinate(packed_event_type event, event_type result, precision_type precision)
{
  unsigned long temp;
  size_type size;
  int32 temp32;
  int64 temp64;
  int96 temp96;

  /* PRECISION_TO_SIZE(precision,size); */
  size = sizes[packed_determinate][precision->precision];

  result->upack.ts_type = size;
  result->upack.base_ts_type = packed_determinate;
  result->upack.modified = FALSE;

  /* Get first word */
  temp = event->determinate.data;
  /* Mask out flag bits */
  if (event->determinate.sign) temp = temp | DETERMINATE_MASK;
  /* Do other words if necessary */
  switch (size) {
    case upack32: 
      CONSTRUCT32(temp,temp32);
      ASSIGN32(temp32,result->upack32.granules);
      break;
    case upack64: 
      CONSTRUCT64(temp,event->word[2],temp64);
      ASSIGN64(temp64,result->upack64.granules);
      break;
    case upack96: 
      CONSTRUCT96(temp,event->word[2],event->word[3],temp96);
      ASSIGN96(temp96,result->upack96.granules);
      break;
    default:
      internal_error("%s %d", "tadt_ unpack_determinate bad size ", size);
    }

  return tadt_ok;
}

#ifdef NOWRELATIVE
/*  
*  Routine:  tadt_unpack_nowrelative
*
*  Description:  Unpacks an nowrelative timestamp into the unpacked format.
*
*  Arguments: event -- (IN): event structure
*             result -- (OUT): result event
*
*  Return Value: Error code
*
*  Errors: none
*  
*  Side Effects: Assignment to passed result structure
*/

static error_type tadt_unpack_nowrelative_fixedspan(packed_event_type event, event_type result, precision_type precision)
{
  event_space_type temp_event, temp_present;
  unsigned long temp;
  size_type size;
  int32 temp32;
  int64 temp64;
  int96 temp96;

  /* PRECISION_TO_SIZE(precision,size); */
  size = sizes[nowrelative][precision->precision];

  temp_event.upack.ts_type = size;
  temp_event.upack.base_ts_type = nowrelative;
  temp_event.upack.modified = FALSE;

  temp = event->nowrelative.data;
  if (event->nowrelative.sign) temp = temp | NOWRELATIVE_MASK;
  switch (size) {
    case upack32: 
      CONSTRUCT32(temp,temp32);
      ASSIGN32(temp32,temp_event.upack32.granules);
      break;
    case upack64: 
      CONSTRUCT64(temp,event->word[2],temp64);
      ASSIGN64(temp64,temp_event.upack64.granules);
      break;
    case upack96: 
      CONSTRUCT96(temp,event->word[2],event->word[3],temp96);
      ASSIGN96(temp96,temp_event.upack96.granules);
      break;
    default:
      internal_error("%s", "bad size");
    }

  tadt_present_as_of_query(FALSE, upack, precision, &temp_present);
  tadt_e_add_fs(&temp_present, (span_type) &temp_event, result);
  return ok;
}
#endif

/*  
*  Routine:  tadt_unpack_special
*
*  Description:  Unpacks an special timestamp into the unpacked format.
*
*  Arguments: event -- (IN): event structure
*             result -- (OUT): result event
*
*  Return Value: Error code
*
*  Errors: none
*  
*  Side Effects: Assignment to passed result structure
*/

static error_type tadt_unpack_special(packed_event_type event, event_type result, precision_type precision)
{

  result->special.which = event->special.which;
  result->upack.ts_type = special; 
  return ok;
}

/*  
*  Routine:  tadt_swizzle_e
*
*  Description:  Swizzles a packed event timestamp into the event format.
*
*  Arguments: event -- (IN): event structure
*             result -- (OUT): result event
*
*  Return Value: Error code
*
*  Errors: none
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_swizzle_e(packed_event_type event, event_type result)
{
  switch (event->generic.ts_type) {
    case packed_determinate:/* do not unpack */
      return tadt_unpack_determinate(event,result,&context->precision);

    case packed_special: 
      return tadt_unpack_special(event,result,&context->precision);

#ifdef INDETERMINATE
    case packed_indeterminate:
      switch (context->resolution) {
        case packed_chunked:
          return tadt_unpack_chunked(event,result,&context->precision);

        case packed_chunked_nonstandard:
          return tadt_unpack_chunked_nonstandard(event,result,&context->precision);

        case packed_general:
          return tadt_unpack_general(event,result,&context->precision);

        case packed_general_nonstandard:
          return tadt_unpack_general_nonstandard(event,result,&context->precision);

        default:
          internal_error("%s", "tadt_swizzle_e: bad indeterminate type ");
      }

#endif

    default:
      internal_error("%s", "tadt_swizzle_e: bad passed ts_type");

    }

  return ok;
}

/*  
*  Routine:  tadt_swizzle_fs
*
*  Description:  Swizzles a span timestamp into the unpacked format.
*
*  Arguments: span -- (IN): span structure
*             result -- (OUT): result span
*
*  Return Value: Error code
*
*  Errors: none
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_swizzle_fs(packed_span_type span, span_type result) {
   /* exploit event, fixed span similarity */
   return tadt_swizzle_e((packed_event_type) span, (event_type) result);
}

/*
*  Routine:  tadt_swizzle_i NEEDS WORK
*
*  Description:  Swizzles an interval timestamp into the unpacked format.
*
*  Arguments: interval -- (IN): interval structure
*             result -- (OUT): result interval
*
*  Return Value: Error code
*
*  Errors: none
* 
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_swizzle_i(interval_type interval, interval_type result) {
   tadt_swizzle_e((packed_event_type) interval->start, result->start);
   tadt_swizzle_e((packed_event_type) interval->terminate, result->terminate);
   return ok;
}

/*  
*  Routine:  tadt_create_indeterminate_e NEEDS WORK
*
*  Description:  Builds an indeterminate event timestamp from the 
*                given information.  
*		 
*  Arguments: lower -- (IN): the lower support
*             upper -- (IN): the upper support
*             pmf -- (IN): the name of the probability mass function
*             result -- (OUT): result event
*
*  Return Value: Error code
*
*  Errors: tadt_bad_support_specified
*          tadt_inconsistent_granularity
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_create_indeterminate_e(event_type lower,
                                               event_type upper,
	                                       pmf_type pmf,
				               event_type result)
{ 
  return tadt_ok;
}

/*  
*  Routine:  tadt_create_indeterminate_now_relative_e NEEDS WORK
*
*  Description:  Builds a indeterminate now-relative event timestamp from the 
*                given information.  
*		 
*  Arguments: limit -- (IN): the upper limit on the indeterminacy
*             relative -- (IN): a fixed or variable span that provides the 
*                               relative in now-relative
*             result -- (OUT): result event
*
*  Return Value: Error code
*
*  Errors: tadt_inconsistent_granularity
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_create_indeterminate_now_relative_e(event_type limit,
	                                            span_type relative,
				                    event_type result)
{ 
  return tadt_ok;
}

/*  
*  Routine:  tadt_create_now_relative_e  NEEDS WORK
*
*  Description:  Builds a now-relative event timestamp from the 
*                given information.  
*		 
*  Arguments: relative -- (IN): a fixed, variable, or indeterminate span 
*                               that provides the relative in now-relative
*             result -- (OUT): result event
*
*  Return Value: Error code
*
*  Errors: none
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_create_now_relative_e(span_type relative,
				      event_type result)
{ 
  return tadt_ok;
}

/*  
*  Routine:  tadt_create_special_e  NEEDS WORK
*
*  Description:  Builds a special event timestamp from the 
*                given information.  
*		 
*  Arguments: which -- (IN): which special guy is this?
*             result -- (OUT): result event
*
*  Return Value: Error code
*
*  Errors: none
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_create_special_e(int which, event_type result)
{ 
  return tadt_ok;
}

/*  
*  Routine:  tadt_create_indeterminate_fs  NEEDS WORK
*
*  Description:  Builds a indeterminate span timestamp from the 
*                given information.  
*		 
*  Arguments: lower -- (IN): the lower support
*             upper -- (IN): the upper support
*             pmf -- (IN): the name of the probability mass function
*             result -- (OUT): result span
*
*  Return Value: Error code
*
*  Errors: tadt_bad_support_specified
*          tadt_inconsistent_granularity
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_create_indeterminate_fs(span_type lower,
                                        span_type upper,
	                                pmf_type pmf,
				        span_type result)
{ 
  return tadt_ok;
}

/*  
*  Routine:  tadt_create_indeterminate_now_relative_fs  NEEDS WORK
*
*  Description:  Builds a indeterminate now-relative span timestamp from the 
*                given information.  
*		 
*  Arguments: limit -- (IN): the upper limit on the indeterminacy
*             relative -- (IN): a fixed or variable span that provides the 
*                               relative in now-relative
*             result -- (OUT): result span
*
*  Return Value: Error code
*
*  Errors: tadt_bad_suppport_specified
*          tadt_inconsistent_granularity
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_create_indeterminate_now_relative_fs(span_type limit,
	                                            span_type relative,
				                    span_type result)
{ 
  return tadt_ok;
}

/*  
*  Routine:  tadt_create_now_relative_fs  NEEDS WORK
*
*  Description:  Builds a now-relative span timestamp from the 
*                given information.  
*		 
*  Arguments: relative -- (IN): a fixed, variable, or indeterminate span 
*                               that provides the relative in now-relative
*             result -- (OUT): result span
*
*  Return Value: Error code
*
*  Errors: none
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_create_now_relative_fs(span_type relative,
				      span_type result)
{ 
  return tadt_ok;
}

/*  
*  Routine:  tadt_create_special_fs  NEEDS WORK
*
*  Description:  Builds a special span timestamp from the 
*                given information.  
*		 
*  Arguments: which -- (IN): which special guy is this?
*             result -- (OUT): result span
*
*  Return Value: Error code
*
*  Errors: none
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_create_special_fs(int which, span_type result)
{ 
  return tadt_ok;
}

/*  
*  Routine:  tadt_deconstruct_e  NEEDS WORK
*
*  Description:  Deconstructs an event
*		 
*  Arguments: event -- (IN): event to deconstruct
*             result -- (OUT): result event
*
*  Return Value: Error code
*
*  Errors: none
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_deconstruct_e(event_type event, deconstructed_event_type result)
{ 
  return tadt_ok;
}

/*  
*  Routine:  tadt_deconstruct_fs  NEEDS WORK
*
*  Description:  Deconstructs a fixed span
*		 
*  Arguments: span -- (IN): span to deconstruct
*             result -- (OUT): result span
*
*  Return Value: Error code
*
*  Errors: none
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_deconstruct_fs(span_type span, deconstructed_span_type result)
{ 
  return tadt_ok;
}

/*  
*  Routine:  tadt_deconstruct_i  NEEDS WORK
*
*  Description:  Deconstructs an interval into two deconstructed events
*		 
*  Arguments: interval -- (IN): interval to deconstruct
*             result_start -- (OUT): result event
*             result_finish -- (OUT): result event
*
*  Return Value: Error code
*
*  Errors: none
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_deconstruct_i(span_type span, deconstructed_event_type result_start, deconstructed_event_type result_finish)
{ 
  return tadt_ok;
}

/*  
*  Routine:  tadt_create_e  NEEDS WORK
*
*  Description:  Builds an event timestamp from the given chronon information.
*		 
*
*  Arguments: granularity -- (IN): granularity
*             size -- (IN): size of the thing
*             granules    -- (IN): number of granules
*             result -- (OUT): result event
*
*  Return Value: Error code
*
*  Errors: none
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_create_e(granularity_type granularity, int size, polymorphic_int_type granules, event_type result)
{ 
/* Move this to unswizzle 
  switch (seconds->resolution) {
    case determinate:
      switch (seconds->size) {
	case SIZE32: ASSIGN32(seconds->seconds32,(int32 *)result); break;
	case SIZE64: ASSIGN64(seconds->seconds64,(int64 *)result); break;
	case SIZE96: ASSIGN96(seconds->seconds96,(int96 *)result); break;
        default: internal_error("%s", "bad size");
	}
      result->determinate.ts_type = determinate;
      break;
*/
      result->upack.modified = FALSE;
      result->upack.base_ts_type = packed_determinate;
      switch (size) {
	case SIZE32: 
           result->upack.ts_type = upack32;
           ASSIGN32(granules->data32,result->upack32.granules); 
           break;
	case SIZE64: 
           result->upack.ts_type = upack64;
           ASSIGN64(granules->data64,result->upack64.granules); 
           break;
	case SIZE96: 
           result->upack.ts_type = upack96;
           ASSIGN96(granules->data96,result->upack96.granules); 
           break;
        default: internal_error("%s", "bad size in tadt_create_e ");
	}
/* Move to special 
      break;

    case special:
	  result->special.which = seconds->special;
          result->special.ts_type = special;
      break;

*/

  return ok;
}

/*  
*  Routine:  tadt_create_i
*
*  Description:  Builds an interval timestamp from the given events.  Assumes
*     storage for the interval has been pre-allocated.  Also tests the events
*     to make sure that it doesn`t build a bad interval.
*
*  Arguments: event1 -- (IN): starting event
*             event2 -- (IN): terminating event
*             result -- (OUT): result interval
*
*  Return Value: Error code
*
*  Errors: out_of_range
*          bad_interval
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_create_i(event_type event1, event_type event2, interval_type result){
  if (tadt_e_precedes_e(event2, event1)) return bad_interval;
  tadt_copy_e(event1, result->start);
  tadt_copy_e(event2, result->terminate);
  return ok;
}

/*  
*  Routine:  tadt_create_fs
*
*  Description:  Builds a fixed span from the given chronon information.
*
*  Arguments: granularity -- (IN): granularity
*             size -- (IN): size of the span to create
*             granules    -- (IN): number of granules
*             result -- (OUT): result span
*
*  Return Value: Error code
*
*  Errors: none
*  
*  Side Effects: Assignment to passed result structure
*/

error_type tadt_create_fs(granularity_type granularity, int size, polymorphic_int_type granules, span_type result)
{ 
  /* exploit fixed-span, event similarity */
  return tadt_create_e(granularity, size, granules, (event_type) result);  
}

/* 
*  Routine:  tadt_create_vs  THIS ROUTINE OUT
*
*  Description:  Builds a variable span from the given information.
*
*  Arguments: cal_sys_id -- (IN):  calendar system identifier
*             data -- (IN):  the calendar interpreted data
*             result -- (OUT): resulting variable span
*
*  Return Value: Error code
*
*  Errors: None
* 
*  Side Effects: Assignment to passed result structure
*/

/*
error_type tadt_create_vs(int cal_sys_id, polymorphic_int_type data, span_type result
) {
}
*/

/*  
*  Routine:  tadt_extract_vs  THIS ROUTINE OUT
*
*  Description:  Extract the variable span information from the given vspan.
*
*  Arguments: vs -- (IN):  variable span time-stamp
*             result -- (OUT): resulting variable span information structure
*
*  Return Value: Error code
*
*  Errors: None
*  
*  Side Effects: Assignment to passed result structure
*/

/*
error_type tadt_extract_vs(span_type vs, vs_type result) {
  result->cal_id = vs->vspan.cal_sys;
  result->data.word[0] = vs->vspan.dataone;
  result->data.word[1] = vs->vspan.dataone;
  return ok;
}
*/

/*  
*  Routine:  tadt_is_vs  THIS ROUTINE OUT
*
*  Description:  Determine if a span is a variable span.
*
*  Arguments: span -- (IN):  the timestamp to be tested
*
*  Return Value: TRUE if span is a variable span
* 		 FALSE otherwise
*  Errors: None
*  
*  Side Effects: None
*/

/*
bool tadt_is_vs(span_type span) {
  return (span->generic.ts_type == special && span->special.vspan);
}
*/

/*  
*  Routine:  tadt_can_fit_ts THIS ROUTINE OUT
*
*  Description:  Determine if there is enough space allocated to 
*                store a time-stamp.  Time-stamp sizes can be
*                determined using the sizeof function.
*
*  Arguments: ts -- (IN):  the timestamp to be tested 
*             size -- (IN):  how much space is available
*
*  Return Value: TRUE if there is enough space 
* 		 FALSE otherwise
*  Errors: None
*  
*  Side Effects: None
*/

/*
bool tadt_can_fit_ts(timestamp_type ts, int size) {
  return FALSE; 
}
*/

/*
*  Routine:  tadt_fs_to_seconds THIS ROUTINE OUT
*
*  Description:  Break a fixed span up into its unpacked form (seconds struct)
*
*  Arguments: fs -- (IN):  the fixed span to be converted
*             seconds -- (OUT):  the resulting seconds structure
*
*  Return Value: Error code
* 
*  Errors: None
* 
*  Side Effects: Assignment to passed seconds structure
*/

/*
error_type tadt_fs_to_seconds(span_type fs, seconds_type seconds) {

  return tadt_e_to_seconds((event_type) fs, seconds);
}
*/

/*  
*  Routine:  tadt_e_to_seconds THIS ROUTINE OUT
*
*  Description:  Break an event up into its unpacked form (seconds structure)
*
*  Arguments: event -- (IN):  the event to be converted
*             seconds -- (OUT):  the resulting seconds structure
*
*  Return Value: Error code
*  
*  Errors: None
*  
*  Side Effects: Assignment to passed seconds structure
*/

/*
error_type tadt_e_to_seconds(event_type event, seconds_type seconds) {
  event_space_type result;

  seconds->resolution = event->generic.ts_type;
  switch (event->generic.ts_type) {

    case determinate:
      seconds->special = 0;
      tadt_swizzle_e(event,&result);
      switch (result.upack.ts_type) {
	case upack32: ASSIGN32(result.upack32.granules,seconds->seconds32); break;
	case upack64: ASSIGN64(result.upack64.granules,seconds->seconds64); break;
	case upack96: ASSIGN96(result.upack96.granules,seconds->seconds96); break;
        default: internal_error("%s", "bad size");
	}
      return ok;

    case upack32:
      seconds->special = 0;
      ASSIGN32(event->upack32.granules,seconds->seconds32); break;
      return ok;

    case upack64:
      seconds->special = 0;
      ASSIGN64(event->upack64.granules,seconds->seconds64); break;
      return ok;

    case upack96:
      seconds->special = 0;
      ASSIGN96(event->upack96.granules,seconds->seconds96); break;
      return ok;

    case special:
      seconds->special = event->special.which;
      return ok;

    default:
      internal_error("%s", "Bad flags");

    }
  return ok;
}
*/

/*  
*  Routine:  tadt_seconds_to_e THIS ROUTINE OUT
*
*  Description:  Convert an unpacked event into its packed form
*
*  Arguments: seconds -- (IN):  the seconds structure to be converted
*             e -- (OUT):  the resulting event
*
*  Return Value: Error code
*  
*  Errors: None
*  
*  Side Effects: Assignment to passed seconds structure
*/

/*
error_type tadt_seconds_to_e(seconds_type seconds, event_type event) {

  return tadt_create_e(seconds, event);

}
*/
